@@ -31,14 +31,19 @@
 #if JAPANESE_SUPPORT
 #include "Japan12CMapInfo.h"
 #endif
-#if CHINESE_SUPPORT
+#if CHINESE_GB_SUPPORT
 #include "GB12CMapInfo.h"
 #endif
+#if CHINESE_CNS_SUPPORT
+#include "CNS13CMapInfo.h"
+#endif
 
 //------------------------------------------------------------------------
 
+extern "C" {
 static int CDECL cmpWidthExcep(const void *w1, const void *w2);
 static int CDECL cmpWidthExcepV(const void *w1, const void *w2);
+}
 
 //------------------------------------------------------------------------
 
@@ -61,7 +66,7 @@ static Gushort *defCharWidths[12] = {
 // GfxFont
 //------------------------------------------------------------------------
 
-GfxFont::GfxFont(char *tag1, Ref id1, Dict *fontDict) {
+GfxFont::GfxFont(char *tagA, Ref idA, Dict *fontDict) {
   BuiltinFont *builtinFont;
   Object obj1, obj2, obj3, obj4;
   int missingWidth;
@@ -69,26 +74,8 @@ GfxFont::GfxFont(char *tag1, Ref id1, Dict *fontDict) {
   int i;
 
   // get font tag and ID
-  tag = new GString(tag1);
-  id = id1;
-
-  // get base font name
-  name = NULL;
-  fontDict->lookup("BaseFont", &obj1);
-  if (obj1.isName())
-    name = new GString(obj1.getName());
-  obj1.free();
-
-  // is it a built-in font?
-  builtinFont = NULL;
-  if (name) {
-    for (i = 0; i < numBuiltinFonts; ++i) {
-      if (!strcmp(builtinFonts[i].name, name->getCString())) {
-	builtinFont = &builtinFonts[i];
-	break;
-      }
-    }
-  }
+  tag = new GString(tagA);
+  id = idA;
 
   // get font type
   type = fontUnknownType;
@@ -106,42 +93,46 @@ GfxFont::GfxFont(char *tag1, Ref id1, Dict *fontDict) {
   obj1.free();
   is16 = gFalse;
 
-  // assume Times-Roman by default (for substitution purposes)
-  flags = fontSerif;
+  // get base font name
+  name = NULL;
+  fontDict->lookup("BaseFont", &obj1);
+  if (obj1.isName())
+    name = new GString(obj1.getName());
+  obj1.free();
 
   // Newer Adobe tools are using Base14-compatible TrueType fonts
   // without embedding them, so munge the names into the equivalent
   // PostScript names.  This is a kludge -- it would be nice if Adobe
   // followed their own spec.
-  if (type == fontTrueType) {
+  if (type == fontTrueType && name) {
     p = name->getCString();
     name2 = NULL;
     if (!strncmp(p, "Arial", 5)) {
       if (!strcmp(p+5, ",Bold")) {
 	name2 = "Helvetica-Bold";
-      } else if (!strcmp(p+5, "Italic")) {
+      } else if (!strcmp(p+5, ",Italic")) {
 	name2 = "Helvetica-Oblique";
-      } else if (!strcmp(p+5, "BoldItalic")) {
+      } else if (!strcmp(p+5, ",BoldItalic")) {
 	name2 = "Helvetica-BoldOblique";
       } else {
 	name2 = "Helvetica";
       }
     } else if (!strncmp(p, "TimesNewRoman", 13)) {
-      if (!strcmp(p+5, ",Bold")) {
+      if (!strcmp(p+13, ",Bold")) {
 	name2 = "Times-Bold";
-      } else if (!strcmp(p+5, "Italic")) {
+      } else if (!strcmp(p+13, ",Italic")) {
 	name2 = "Times-Italic";
-      } else if (!strcmp(p+5, "BoldItalic")) {
+      } else if (!strcmp(p+13, ",BoldItalic")) {
 	name2 = "Times-BoldItalic";
       } else {
 	name2 = "Times-Roman";
       }
     } else if (!strncmp(p, "CourierNew", 10)) {
-      if (!strcmp(p+5, ",Bold")) {
+      if (!strcmp(p+10, ",Bold")) {
 	name2 = "Courier-Bold";
-      } else if (!strcmp(p+5, "Italic")) {
+      } else if (!strcmp(p+10, ",Italic")) {
 	name2 = "Courier-Oblique";
-      } else if (!strcmp(p+5, "BoldItalic")) {
+      } else if (!strcmp(p+10, ",BoldItalic")) {
 	name2 = "Courier-BoldOblique";
       } else {
 	name2 = "Courier";
@@ -153,12 +144,53 @@ GfxFont::GfxFont(char *tag1, Ref id1, Dict *fontDict) {
     }
   }
 
+  // is it a built-in font?
+  builtinFont = NULL;
+  if (name) {
+    for (i = 0; i < numBuiltinFonts; ++i) {
+      if (!strcmp(builtinFonts[i].name, name->getCString())) {
+	builtinFont = &builtinFonts[i];
+	break;
+      }
+    }
+  }
+
+  // assume Times-Roman by default (for substitution purposes)
+  flags = fontSerif;
+
+  // default ascent/descent values
+  if (builtinFont) {
+    ascent = 0.001 * builtinFont->ascent;
+    descent = 0.001 * builtinFont->descent;
+  } else {
+    ascent = 0.95;
+    descent = -0.35;
+  }
+
   // get info from font descriptor
   embFontName = NULL;
   embFontID.num = -1;
   embFontID.gen = -1;
   missingWidth = 0;
-  fontDict->lookup("FontDescriptor", &obj1);
+  if (type == fontType0) {
+    fontDict->lookup("DescendantFonts", &obj2);
+    if (obj2.isArray()) {
+      obj2.arrayGet(0, &obj3);
+      if (obj3.isDict()) {
+	obj3.dictLookup("FontDescriptor", &obj1);
+      } else {
+	error(-1, "Bad descendant font in Type 0 font");
+	obj1.initNull();
+      }
+      obj3.free();
+    } else {
+      error(-1, "Missing DescendantFonts entry in Type 0 font");
+      obj1.initNull();
+    }
+    obj2.free();
+  } else {
+    fontDict->lookup("FontDescriptor", &obj1);
+  }
   if (obj1.isDict()) {
 
     // get flags
@@ -216,6 +248,18 @@ GfxFont::GfxFont(char *tag1, Ref id1, Dict *fontDict) {
       missingWidth = obj2.getInt();
     }
     obj2.free();
+
+    // get Ascent and Descent
+    obj1.dictLookup("Ascent", &obj2);
+    if (obj2.isNum()) {
+      ascent = 0.001 * obj2.getNum();
+    }
+    obj2.free();
+    obj1.dictLookup("Descent", &obj2);
+    if (obj2.isNum()) {
+      descent = 0.001 * obj2.getNum();
+    }
+    obj2.free();
   }
   obj1.free();
 
@@ -426,24 +470,29 @@ void GfxFont::getEncAndWidths(Dict *fontDict, BuiltinFont *builtinFont,
   obj1.free();
 
   // check embedded or external font file for base encoding
-  if ((type == fontType1 || type == fontType1C) &&
+  if ((type == fontType1 || type == fontType1C || type == fontTrueType) &&
       (extFontFile || embFontID.num >= 0)) {
-    if (extFontFile)
+    if (extFontFile) {
       buf = readExtFontFile(&len);
-    else
+    } else {
       buf = readEmbFontFile(&len);
+    }
     if (buf) {
-      if (type == fontType1)
+      if (type == fontType1) {
 	fontFile = new Type1FontFile(buf, len);
-      else
+      } else if (type == fontType1C) {
 	fontFile = new Type1CFontFile(buf, len);
+      } else {
+	fontFile = new TrueTypeFontFile(buf, len);
+      }
       if (fontFile->getName()) {
 	if (embFontName)
 	  delete embFontName;
 	embFontName = new GString(fontFile->getName());
       }
-      if (!encoding)
+      if (!encoding) {
 	encoding = fontFile->getEncoding(gTrue);
+      }
       delete fontFile;
       gfree(buf);
     }
@@ -564,6 +613,7 @@ char *GfxFont::readEmbFontFile(int *len) {
     buf[i++] = c;
   }
   *len = i;
+  str->close();
 
   obj2.free();
   obj1.free();
@@ -586,64 +636,67 @@ void GfxFont::makeWidths(Dict *fontDict, FontEncoding *builtinEncoding,
     widths[code] = missingWidth * 0.001;
   }
 
+  // use widths from font dict, if present
+  fontDict->lookup("FirstChar", &obj1);
+  firstChar = obj1.isInt() ? obj1.getInt() : 0;
+  obj1.free();
+  fontDict->lookup("LastChar", &obj1);
+  lastChar = obj1.isInt() ? obj1.getInt() : 255;
+  obj1.free();
+  if (type == fontType3)
+    mult = fontMat[0];
+  else
+    mult = 0.001;
+  fontDict->lookup("Widths", &obj1);
+  if (obj1.isArray()) {
+    for (code = firstChar; code <= lastChar; ++code) {
+      obj1.arrayGet(code - firstChar, &obj2);
+      if (obj2.isNum())
+	widths[code] = obj2.getNum() * mult;
+      obj2.free();
+    }
+
   // use widths from built-in font
-  if (builtinEncoding) {
+  } else if (builtinEncoding) {
     code2 = 0; // to make gcc happy
     for (code = 0; code < 256; ++code) {
       if ((charName = encoding->getCharName(code)) &&
-	  (code2 = builtinEncoding->getCharCode(charName)) >= 0)
+	  (code2 = builtinEncoding->getCharCode(charName)) >= 0) {
 	widths[code] = builtinWidths[code2] * 0.001;
+      } else if (code == 32) {
+	// this is a kludge for broken PDF files that encode char 32
+	// as .notdef
+	widths[code] = builtinWidths[' '] * 0.001;
+      }
     }
 
-  // get widths from font dict
+  // couldn't find widths -- use defaults 
   } else {
-    fontDict->lookup("FirstChar", &obj1);
-    firstChar = obj1.isInt() ? obj1.getInt() : 0;
-    obj1.free();
-    fontDict->lookup("LastChar", &obj1);
-    lastChar = obj1.isInt() ? obj1.getInt() : 255;
-    obj1.free();
-    if (type == fontType3)
-      mult = fontMat[0];
-    else
-      mult = 0.001;
-    fontDict->lookup("Widths", &obj1);
-    if (obj1.isArray()) {
-      for (code = firstChar; code <= lastChar; ++code) {
-	obj1.arrayGet(code - firstChar, &obj2);
-	if (obj2.isNum())
-	  widths[code] = obj2.getNum() * mult;
-	obj2.free();
-      }
-    } else {
-
-      // couldn't find widths -- use defaults 
 #if 0 //~
-      //~ certain PDF generators apparently don't include widths
-      //~ for Arial and TimesNewRoman -- and this error message
-      //~ is a nuisance
-      error(-1, "No character widths resource for non-builtin font");
+    //~ certain PDF generators apparently don't include widths
+    //~ for Arial and TimesNewRoman -- and this error message
+    //~ is a nuisance
+    error(-1, "No character widths resource for non-builtin font");
 #endif
-      if (isFixedWidth())
-	index = 0;
-      else if (isSerif())
-	index = 8;
-      else
-	index = 4;
-      if (isBold())
-	index += 2;
-      if (isItalic())
-	index += 1;
-      defWidths = defCharWidths[index];
-      code2 = 0; // to make gcc happy
-      for (code = 0; code < 256; ++code) {
-	if ((charName = encoding->getCharName(code)) &&
-	    (code2 = standardEncoding.getCharCode(charName)) >= 0)
-	  widths[code] = defWidths[code2] * 0.001;
-      }
+    if (isFixedWidth())
+      index = 0;
+    else if (isSerif())
+      index = 8;
+    else
+      index = 4;
+    if (isBold())
+      index += 2;
+    if (isItalic())
+      index += 1;
+    defWidths = defCharWidths[index];
+    code2 = 0; // to make gcc happy
+    for (code = 0; code < 256; ++code) {
+      if ((charName = encoding->getCharName(code)) &&
+	  (code2 = standardEncoding.getCharCode(charName)) >= 0)
+	widths[code] = defWidths[code2] * 0.001;
     }
-    obj1.free();
   }
+  obj1.free();
 }
 
 void GfxFont::getType0EncAndWidths(Dict *fontDict) {
@@ -686,15 +739,24 @@ void GfxFont::getType0EncAndWidths(Dict *fontDict) {
 #endif
     } else if (obj4.getString()->cmp("Adobe") == 0 &&
 	       obj5.getString()->cmp("GB1") == 0) {
-#if CHINESE_SUPPORT
+#if CHINESE_GB_SUPPORT
       is16 = gTrue;
       enc16.charSet = font16AdobeGB12;
 #else
-      error(-1, "Xpdf was compiled without Chinese font support");
+      error(-1, "Xpdf was compiled without Chinese GB font support");
+      goto err4;
+#endif
+    } else if (obj4.getString()->cmp("Adobe") == 0 &&
+	       obj5.getString()->cmp("CNS1") == 0) {
+#if CHINESE_CNS_SUPPORT
+      is16 = gTrue;
+      enc16.charSet = font16AdobeCNS13;
+#else
+      error(-1, "Xpdf was compiled without Chinese CNS font support");
       goto err4;
 #endif
     } else {
-      error(-1, "Uknown Type 0 character set: %s-%s",
+      error(-1, "Unknown Type 0 character set: %s-%s",
 	    obj4.getString()->getCString(), obj5.getString()->getCString());
       goto err4;
     }
@@ -907,7 +969,7 @@ void GfxFont::getType0EncAndWidths(Dict *fontDict) {
     enc16.enc = gfxJapan12Tab[i].enc;
   }
 #endif
-#if CHINESE_SUPPORT
+#if CHINESE_GB_SUPPORT
   if (enc16.charSet == font16AdobeGB12) {
     for (i = 0; gfxGB12Tab[i].name; ++i) {
       if (!strcmp(obj1.getName(), gfxGB12Tab[i].name))
@@ -920,6 +982,20 @@ void GfxFont::getType0EncAndWidths(Dict *fontDict) {
     }
     enc16.enc = gfxGB12Tab[i].enc;
   }
+#endif
+#if CHINESE_CNS_SUPPORT
+  if (enc16.charSet == font16AdobeCNS13) {
+    for (i = 0; gfxCNS13Tab[i].name; ++i) {
+      if (!strcmp(obj1.getName(), gfxCNS13Tab[i].name))
+	break;
+    }
+    if (!gfxCNS13Tab[i].name) {
+      error(-1, "Unknown encoding '%s' for Adobe-CNS1-3 font",
+	    obj1.getName());
+      goto err1;
+    }
+    enc16.enc = gfxCNS13Tab[i].enc;
+  }
 #endif
   obj1.free();
 
@@ -939,6 +1015,7 @@ void GfxFont::getType0EncAndWidths(Dict *fontDict) {
   makeWidths(fontDict, NULL, NULL, 0);
 }
 
+extern "C" {
 static int CDECL cmpWidthExcep(const void *w1, const void *w2) {
   return ((GfxFontWidthExcep *)w1)->first - ((GfxFontWidthExcep *)w2)->first;
 }
@@ -946,6 +1023,7 @@ static int CDECL cmpWidthExcep(const void *w1, const void *w2) {
 static int CDECL cmpWidthExcepV(const void *w1, const void *w2) {
   return ((GfxFontWidthExcepV *)w1)->first - ((GfxFontWidthExcepV *)w2)->first;
 }
+}
 
 //------------------------------------------------------------------------
 // GfxFontDict
