@@ -102,8 +102,8 @@ enum XpdfMenuItem {
   menuOpen,
   menuReload,
   menuSavePDF,
-  menuRotateLeft,
-  menuRotateRight,
+  menuRotateCCW,
+  menuRotateCW,
   menuQuit
 };
 
@@ -133,8 +133,10 @@ static void mouseDragCbk(LTKWidget *widget, int widgetNum,
 // button callbacks
 static void nextPageCbk(LTKWidget *button, int n, GBool on);
 static void nextTenPageCbk(LTKWidget *button, int n, GBool on);
+static void gotoNextPage(int inc, GBool top);
 static void prevPageCbk(LTKWidget *button, int n, GBool on);
 static void prevTenPageCbk(LTKWidget *button, int n, GBool on);
+static void gotoPrevPage(int dec, GBool top, GBool bottom);
 static void backCbk(LTKWidget *button, int n, GBool on);
 static void forwardCbk(LTKWidget *button, int n, GBool on);
 static void pageNumCbk(LTKWidget *textIn, int n, GString *text);
@@ -199,6 +201,7 @@ static XrmOptionDescRec opts[] = {
   {"-g",             ".geometry",        XrmoptionSepArg,  NULL},
   {"-font",          ".font",            XrmoptionSepArg,  NULL},
   {"-fn",            ".font",            XrmoptionSepArg,  NULL},
+  {"-title",         ".title",           XrmoptionSepArg,  NULL},
   {"-cmap",          ".installCmap",     XrmoptionNoArg,   (XPointer)"on"},
   {"-rgb",           ".rgbCubeSize",     XrmoptionSepArg,  NULL},
   {"-papercolor",    ".paperColor",      XrmoptionSepArg,  NULL},
@@ -208,7 +211,7 @@ static XrmOptionDescRec opts[] = {
 #if HAVE_T1LIB_H
   {"-t1lib",         ".t1libControl",    XrmoptionSepArg,  NULL},
 #endif
-#if HAVE_FREETYPE_FREETYPE_H
+#if HAVE_FREETYPE_FREETYPE_H | HAVE_FREETYPE_H
   {"-freetype",      ".freeTypeControl", XrmoptionSepArg,  NULL},
 #endif
   {"-z",             ".initialZoom",     XrmoptionSepArg,  NULL},
@@ -225,6 +228,7 @@ static char remoteName[100] = "xpdf_";
 static GBool doRemoteRaise = gFalse;
 static GBool doRemoteQuit = gFalse;
 static GBool viKeys = gFalse;
+static char ownerPassword[33] = "";
 static char userPassword[33] = "";
 static GBool fullScreen = gFalse;
 
@@ -235,6 +239,8 @@ static ArgDesc argDesc[] = {
    "initial window geometry"},
   {"-geometry",   argStringDummy, NULL,           0,
    "initial window geometry"},
+  {"-title",      argStringDummy, NULL,           0,
+   "window title"},
   {"-remote",     argString,      remoteName + 5, sizeof(remoteName) - 5,
    "start/contact xpdf remote server with specified name"},
   {"-raise",      argFlag,        &doRemoteRaise, 0,
@@ -255,7 +261,7 @@ static ArgDesc argDesc[] = {
   {"-t1lib",      argStringDummy, NULL,           0,
    "t1lib font control: none, plain, low, high"},
 #endif
-#if HAVE_FREETYPE_FREETYPE_H
+#if HAVE_FREETYPE_FREETYPE_H | HAVE_FREETYPE_H
   {"-freetype",   argStringDummy, NULL,           0,
    "FreeType font control: none, plain, aa"},
 #endif
@@ -267,6 +273,8 @@ static ArgDesc argDesc[] = {
    "paper height, in points"},
   {"-level1",     argFlagDummy,   NULL,           0,
    "generate Level 1 PostScript"},
+  {"-opw",        argString,      ownerPassword,  sizeof(ownerPassword),
+   "owner password (for encrypted files)"},
   {"-upw",        argString,      userPassword,   sizeof(userPassword),
    "user password (for encrypted files)"},
   {"-fullscreen", argFlag,        &fullScreen,    0,
@@ -281,6 +289,10 @@ static ArgDesc argDesc[] = {
    "print usage information"},
   {"-help",       argFlag,        &printHelp,     0,
    "print usage information"},
+  {"--help",  argFlag,     &printHelp,     0,
+   "print usage information"},
+  {"-?",      argFlag,     &printHelp,     0,
+   "print usage information"},
   {NULL}
 };
 
@@ -341,6 +353,8 @@ static GString *fileReqDir;	// current directory for file requesters
 
 static GString *urlCommand;	// command to execute for URI links
 
+static GString *windowTitle;	// window title string
+
 static LTKApp *app;
 static Display *display;
 static LTKWindow *win;
@@ -356,6 +370,7 @@ static LTKWindow *psDialog;
 static LTKWindow *openDialog;
 static LTKWindow *saveDialog;
 static LTKWindow *findWin;
+static LTKTextIn *findTextIn;
 static Atom remoteAtom;
 static GC selectGC;
 
@@ -390,6 +405,7 @@ int main(int argc, char *argv[]) {
   int pg;
   int x, y;
   Guint width, height;
+  double width1, height1;
   GString *zoomStr;
   GBool ok;
   char s[20];
@@ -415,10 +431,10 @@ int main(int argc, char *argv[]) {
   errorInit();
 
   // read config file
-  initParams(xpdfConfigFile);
+  initParams(xpdfUserConfigFile, xpdfSysConfigFile);
 
   // create LTKApp (and parse X-related args)
-  app = new LTKApp("xpdf", opts, &argc, argv);
+  app = new LTKApp("xpdf", "Xpdf", opts, &argc, argv);
   app->setKillCbk(&killCbk);
   display = app->getDisplay();
 
@@ -526,6 +542,7 @@ int main(int argc, char *argv[]) {
   paperHeight = app->getIntResource("psPaperHeight", defPaperHeight);
   psOutLevel1 = app->getBoolResource("psLevel1", gFalse);
   urlCommand = app->getStringResource("urlCommand", NULL);
+  windowTitle = app->getStringResource("title", NULL);
   installCmap = app->getBoolResource("installCmap", gFalse);
   if (installCmap)
     win->setInstallCmap(gTrue);
@@ -541,7 +558,7 @@ int main(int argc, char *argv[]) {
 #if HAVE_T1LIB_H
   t1libControl = app->getStringResource("t1libControl", "low");
 #endif
-#if HAVE_FREETYPE_FREETYPE_H
+#if HAVE_FREETYPE_FREETYPE_H | HAVE_FREETYPE_H
   freeTypeControl = app->getStringResource("freeTypeControl", "aa");
 #endif
   t1Courier = app->getStringResource("t1Courier", NULL);
@@ -562,8 +579,11 @@ int main(int argc, char *argv[]) {
 #if JAPANESE_SUPPORT
   japan12Font = app->getStringResource("japaneseFont", NULL);
 #endif
-#if CHINESE_SUPPORT
-  gb12Font = app->getStringResource("chineseFont", NULL);
+#if CHINESE_GB_SUPPORT
+  gb12Font = app->getStringResource("chineseGBFont", NULL);
+#endif
+#if CHINESE_CNS_SUPPORT
+  cns13Font = app->getStringResource("chineseCNSFont", NULL);
 #endif
   if (fullScreen) {
     zoom = zoomPage;
@@ -599,23 +619,25 @@ int main(int argc, char *argv[]) {
     app->getGeometryResource("geometry", &x, &y, &width, &height);
     if (width == 0 || height == 0) {
       if (!doc || doc->getNumPages() == 0) {
-	width = 612;
-	height = 792;
+	width1 = 612;
+	height1 = 792;
       } else if (doc->getPageRotate(pg) == 90 ||
 		 doc->getPageRotate(pg) == 270) {
-	width = (int)(doc->getPageHeight(pg) + 0.5);
-	height = (int)(doc->getPageWidth(pg) + 0.5);
+	width1 = doc->getPageHeight(pg);
+	height1 = doc->getPageWidth(pg);
       } else {
-	width = (int)(doc->getPageWidth(pg) + 0.5);
-	height = (int)(doc->getPageHeight(pg) + 0.5);
+	width1 = doc->getPageWidth(pg);
+	height1 = doc->getPageHeight(pg);
       }
       if (zoom == zoomPage || zoom == zoomWidth) {
-	width = (width * zoomDPI[defZoom - minZoom]) / 72 + 28;
-	height = (height * zoomDPI[defZoom - minZoom]) / 72 + 56;
+	width = (int)((width1 * zoomDPI[defZoom - minZoom]) / 72 + 0.5);
+	height = (int)((height1 * zoomDPI[defZoom - minZoom]) / 72 + 0.5);
       } else {
-	width = (width * zoomDPI[zoom - minZoom]) / 72 + 28;
-	height = (height * zoomDPI[zoom - minZoom]) / 72 + 56;
+	width = (int)((width1 * zoomDPI[zoom - minZoom]) / 72 + 0.5);
+	height = (int)((height1 * zoomDPI[zoom - minZoom]) / 72 + 0.5);
       }
+      width += 28;
+      height += 56;
       if (width > (Guint)app->getDisplayWidth() - 100) {
 	width = app->getDisplayWidth() - 100;
       }
@@ -630,7 +652,9 @@ int main(int argc, char *argv[]) {
     sprintf(s, "of %d", doc ? doc->getNumPages() : 0);
     numPagesLabel->setText(s);
   }
-  if (name) {
+  if (windowTitle) {
+    title = windowTitle->copy();
+  } else if (name) {
     title = new GString("xpdf: ");
     title->append(name);
   } else {
@@ -677,75 +701,110 @@ int main(int argc, char *argv[]) {
 
  done2:
   // free stuff
-  if (out)
+  if (out) {
     delete out;
-  if (win)
+  }
+  if (win) {
     delete win;
-  if (aboutWin)
+  }
+  if (aboutWin) {
     delete aboutWin;
-  if (findWin)
+  }
+  if (findWin) {
     delete findWin;
-  if (app)
+  }
+  if (app) {
     delete app;
-  if (doc)
+  }
+  if (doc) {
     delete doc;
-  if (psFileName)
+  }
+  if (psFileName) {
     delete psFileName;
-  if (defPSFileName)
+  }
+  if (defPSFileName) {
     delete defPSFileName;
-  if (fileReqDir)
+  }
+  if (fileReqDir) {
     delete fileReqDir;
-  if (urlCommand)
+  }
+  if (urlCommand) {
     delete urlCommand;
+  }
+  if (windowTitle) {
+    delete windowTitle;
+  }
 #if HAVE_T1LIB_H
-  if (t1libControl)
+  if (t1libControl) {
     delete t1libControl;
+  }
 #endif
-#if HAVE_FREETYPE_FREETYPE_H
-  if (freeTypeControl)
+#if HAVE_FREETYPE_FREETYPE_H | HAVE_FREETYPE_H
+  if (freeTypeControl) {
     delete freeTypeControl;
+  }
 #endif
-  if (t1Courier)
+  if (t1Courier) {
     delete t1Courier;
-  if (t1CourierBold)
+  }
+  if (t1CourierBold) {
     delete t1CourierBold;
-  if (t1CourierBoldOblique)
+  }
+  if (t1CourierBoldOblique) {
     delete t1CourierBoldOblique;
-  if (t1CourierOblique)
+  }
+  if (t1CourierOblique) {
     delete t1CourierOblique;
-  if (t1Helvetica)
+  }
+  if (t1Helvetica) {
     delete t1Helvetica;
-  if (t1HelveticaBold)
+  }
+  if (t1HelveticaBold) {
     delete t1HelveticaBold;
-  if (t1HelveticaBoldOblique)
+  }
+  if (t1HelveticaBoldOblique) {
     delete t1HelveticaBoldOblique;
-  if (t1HelveticaOblique)
+  }
+  if (t1HelveticaOblique) {
     delete t1HelveticaOblique;
-  if (t1Symbol)
+  }
+  if (t1Symbol) {
     delete t1Symbol;
-  if (t1TimesBold)
+  }
+  if (t1TimesBold) {
     delete t1TimesBold;
-  if (t1TimesBoldItalic)
+  }
+  if (t1TimesBoldItalic) {
     delete t1TimesBoldItalic;
-  if (t1TimesItalic)
+  }
+  if (t1TimesItalic) {
     delete t1TimesItalic;
-  if (t1TimesRoman)
+  }
+  if (t1TimesRoman) {
     delete t1TimesRoman;
-  if (t1ZapfDingbats)
+  }
+  if (t1ZapfDingbats) {
     delete t1ZapfDingbats;
+  }
 #if JAPANESE_SUPPORT
   if (japan12Font) {
     delete japan12Font;
   }
 #endif
-#if CHINESE_SUPPORT
+#if CHINESE_GB_SUPPORT
   if (gb12Font) {
     delete gb12Font;
   }
+#endif
+#if CHINESE_CNS_SUPPORT
+  if (cns13Font) {
+    delete cns13Font;
+  }
 #endif
   for (i = 0; i < historySize; ++i) {
-    if (history[i].fileName)
+    if (history[i].fileName) {
       delete history[i].fileName;
+    }
   }
   freeParams();
 
@@ -763,7 +822,7 @@ int main(int argc, char *argv[]) {
 static GBool loadFile(GString *fileName) {
   GString *title;
   PDFDoc *newDoc;
-  GString *userPW;
+  GString *ownerPW, *userPW;
   char s[20];
   char *p;
 
@@ -772,15 +831,23 @@ static GBool loadFile(GString *fileName) {
     win->setBusyCursor(gTrue);
 
   // open PDF file
+  if (ownerPassword[0]) {
+    ownerPW = new GString(ownerPassword);
+  } else {
+    ownerPW = NULL;
+  }
   if (userPassword[0]) {
     userPW = new GString(userPassword);
   } else {
     userPW = NULL;
   }
-  newDoc = new PDFDoc(fileName, userPW);
+  newDoc = new PDFDoc(fileName, ownerPW, userPW);
   if (userPW) {
     delete userPW;
   }
+  if (ownerPW) {
+    delete ownerPW;
+  }
   if (!newDoc->isOk()) {
     delete newDoc;
     if (win)
@@ -820,9 +887,11 @@ static GBool loadFile(GString *fileName) {
 
   // set up title, number-of-pages display; back to normal cursor
   if (win) {
-    title = new GString("xpdf: ");
-    title->append(fileName);
-    win->setTitle(title);
+    if (!windowTitle) {
+      title = new GString("xpdf: ");
+      title->append(fileName);
+      win->setTitle(title);
+    }
     if (!fullScreen) {
       sprintf(s, "of %d", doc->getNumPages());
       numPagesLabel->setText(s);
@@ -892,18 +961,18 @@ static void displayPage(int page1, int zoom1, int rotate1, GBool addToHist) {
     dpi = (hDPI < vDPI) ? hDPI : vDPI;
   } else if (zoom == zoomPage) {
     if (rot == 90 || rot == 270) {
-      hDPI = ((win->getWidth() - 28) / doc->getPageHeight(page)) * 72;
-      vDPI = ((win->getHeight() - 56) / doc->getPageWidth(page)) * 72;
+      hDPI = (canvas->getWidth() / doc->getPageHeight(page)) * 72;
+      vDPI = (canvas->getHeight() / doc->getPageWidth(page)) * 72;
     } else {
-      hDPI = ((win->getWidth() - 28) / doc->getPageWidth(page)) * 72;
-      vDPI = ((win->getHeight() - 56) / doc->getPageHeight(page)) * 72;
+      hDPI = (canvas->getWidth() / doc->getPageWidth(page)) * 72;
+      vDPI = (canvas->getHeight() / doc->getPageHeight(page)) * 72;
     }
     dpi = (hDPI < vDPI) ? hDPI : vDPI;
   } else if (zoom == zoomWidth) {
     if (rot == 90 || rot == 270) {
-      dpi = ((win->getWidth() - 28) / doc->getPageHeight(page)) * 72;
+      dpi = (canvas->getWidth() / doc->getPageHeight(page)) * 72;
     } else {
-      dpi = ((win->getWidth() - 28) / doc->getPageWidth(page)) * 72;
+      dpi = (canvas->getWidth() / doc->getPageWidth(page)) * 72;
     }
   } else {
     dpi = zoomDPI[zoom - minZoom];
@@ -957,17 +1026,17 @@ static void keyPressCbk(LTKWindow *win1, KeySym key, Guint modifiers,
       break;
     case 'N':
     case 'n':
-      nextPageCbk(NULL, 0, gTrue);
+      gotoNextPage(1, !(modifiers & Mod5Mask));
       break;
     case 'P':
     case 'p':
-      prevPageCbk(NULL, 0, gTrue);
+      gotoPrevPage(1, !(modifiers & Mod5Mask), gFalse);
       break;
     case ' ':
       if (fullScreen ||
 	  vScrollbar->getPos() >=
 	    canvas->getRealHeight() - canvas->getHeight()) {
-	nextPageCbk(NULL, 0, gTrue);
+	gotoNextPage(1, gTrue);
       } else {
 	vScrollbar->setPos(vScrollbar->getPos() + canvas->getHeight(),
 			   canvas->getHeight());
@@ -977,12 +1046,9 @@ static void keyPressCbk(LTKWindow *win1, KeySym key, Guint modifiers,
     case '\b':			// bs
     case '\177':		// del
       if (fullScreen) {
-	prevPageCbk(NULL, 0, gTrue);
+	gotoPrevPage(1, gTrue, gFalse);
       } else if (vScrollbar->getPos() == 0) {
-	prevPageCbk(NULL, 0, gTrue);
-	vScrollbar->setPos(canvas->getRealHeight() - canvas->getHeight(),
-			   canvas->getHeight());
-	canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+	gotoPrevPage(1, gFalse, gTrue);
       } else {
 	vScrollbar->setPos(vScrollbar->getPos() - canvas->getHeight(),
 			   canvas->getHeight());
@@ -1025,6 +1091,41 @@ static void keyPressCbk(LTKWindow *win1, KeySym key, Guint modifiers,
 	canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
       }
       break;
+    case '0':
+      if (fullScreen) {
+	break;
+      }
+      zoom = 0;
+      displayPage(page, zoom, rotate, gFalse);
+      break;
+    case '+':
+      if (fullScreen) {
+	break;
+      }
+      if (zoom >= minZoom && zoom < maxZoom) {
+	zoomMenuBtn->setMenuItem(zoomMenu->getItem(zoom + 1 - minZoom));
+      }
+      break;
+    case '-':
+      if (fullScreen) {
+	break;
+      }
+      if (zoom > minZoom && zoom <= maxZoom) {
+	zoomMenuBtn->setMenuItem(zoomMenu->getItem(zoom - 1 - minZoom));
+      }
+      break;
+    case 'z':
+      if (fullScreen) {
+	break;
+      }
+      zoomMenuBtn->setMenuItem(zoomMenu->getItem(maxZoom - minZoom + 2));
+      break;
+    case 'w':
+      if (fullScreen) {
+	break;
+      }
+      zoomMenuBtn->setMenuItem(zoomMenu->getItem(maxZoom - minZoom + 3));
+      break;
     case '\014':		// ^L
       win->redraw();
       displayPage(page, zoom, rotate, gFalse);
@@ -1059,12 +1160,9 @@ static void keyPressCbk(LTKWindow *win1, KeySym key, Guint modifiers,
     case XK_Page_Up:
     case XK_KP_Page_Up:
       if (fullScreen) {
-	prevPageCbk(NULL, 0, gTrue);
+	gotoPrevPage(1, gTrue, gFalse);
       } else if (vScrollbar->getPos() == 0) {
-	prevPageCbk(NULL, 0, gTrue);
-	vScrollbar->setPos(canvas->getRealHeight() - canvas->getHeight(),
-			   canvas->getHeight());
-	canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+	gotoPrevPage(1, gFalse, gTrue);
       } else {
 	vScrollbar->setPos(vScrollbar->getPos() - canvas->getHeight(),
 			   canvas->getHeight());
@@ -1076,7 +1174,7 @@ static void keyPressCbk(LTKWindow *win1, KeySym key, Guint modifiers,
       if (fullScreen ||
 	  vScrollbar->getPos() >=
 	    canvas->getRealHeight() - canvas->getHeight()) {
-	nextPageCbk(NULL, 0, gTrue);
+	gotoNextPage(1, gTrue);
       } else {
 	vScrollbar->setPos(vScrollbar->getPos() + canvas->getHeight(),
 			   canvas->getHeight());
@@ -1133,13 +1231,13 @@ static void menuCbk(LTKMenuItem *item) {
     if (doc)
       mapSaveDialog();
     break;
-  case menuRotateLeft:
+  case menuRotateCCW:
     if (doc) {
       r = (rotate == 0) ? 270 : rotate - 90;
       displayPage(page, zoom, r, gFalse);
     }
     break;
-  case menuRotateRight:
+  case menuRotateCW:
     if (doc) {
       r = (rotate == 270) ? 0 : rotate + 90;
       displayPage(page, zoom, r, gFalse);
@@ -1157,13 +1255,53 @@ static void menuCbk(LTKMenuItem *item) {
 
 static void buttonPressCbk(LTKWidget *canvas1, int n,
 			   int mx, int my, int button, GBool dblClick) {
-  if (!doc || doc->getNumPages() == 0)
+  if (!doc || doc->getNumPages() == 0) {
     return;
-  if (button == 1) {
+  }
+  switch (button) {
+  case 1:
     setSelection(mx, my, mx, my);
-  } else if (!fullScreen && button == 2) {
-    panMX = mx - hScrollbar->getPos();
-    panMY = my - vScrollbar->getPos();
+    break;
+  case 2:
+    if (!fullScreen) {
+      panMX = mx - hScrollbar->getPos();
+      panMY = my - vScrollbar->getPos();
+    }
+    break;
+  case 4: // mouse wheel up
+    if (fullScreen) {
+      gotoPrevPage(1, gTrue, gFalse);
+    } else if (vScrollbar->getPos() == 0) {
+      gotoPrevPage(1, gFalse, gTrue);
+    } else {
+      vScrollbar->setPos(vScrollbar->getPos() - 16, canvas->getHeight());
+      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    }
+    break;
+  case 5: // mouse wheel down
+    if (fullScreen ||
+	vScrollbar->getPos() >=
+	canvas->getRealHeight() - canvas->getHeight()) {
+      gotoNextPage(1, gTrue);
+    } else {
+      vScrollbar->setPos(vScrollbar->getPos() + 16, canvas->getHeight());
+      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    }
+    break;
+  case 6: // second mouse wheel right
+    if (fullScreen) {
+      return;
+    }
+    hScrollbar->setPos(hScrollbar->getPos() + 16, canvas->getWidth());
+    canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    break;
+  case 7: // second mouse wheel left
+    if (fullScreen) {
+      return;
+    }
+    hScrollbar->setPos(hScrollbar->getPos() - 16, canvas->getWidth());
+    canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    break;
   }
 }
 
@@ -1201,6 +1339,7 @@ static void doLink(int mx, int my) {
   GString *namedDest;
   char *s;
   GString *fileName;
+  GString *actionName;
   Ref pageRef;
   int pg;
   double x, y;
@@ -1400,6 +1539,32 @@ static void doLink(int mx, int my) {
       }
       break;
 
+    // Named action
+    case actionNamed:
+      actionName = ((LinkNamed *)action)->getName();
+      if (!actionName->cmp("NextPage")) {
+	gotoNextPage(1, gTrue);
+      } else if (!actionName->cmp("PrevPage")) {
+	gotoPrevPage(1, gTrue, gFalse);
+      } else if (!actionName->cmp("FirstPage")) {
+	if (page != 1) {
+	  displayPage(1, zoom, rotate, gTrue);
+	}
+      } else if (!actionName->cmp("LastPage")) {
+	if (page != doc->getNumPages()) {
+	  displayPage(doc->getNumPages(), zoom, rotate, gTrue);
+	}
+      } else if (!actionName->cmp("GoBack")) {
+	backCbk(NULL, 0, gTrue);
+      } else if (!actionName->cmp("GoForward")) {
+	forwardCbk(NULL, 0, gTrue);
+      } else if (!actionName->cmp("Quit")) {
+	quitCbk(NULL, 0, gTrue);
+      } else {
+	error(-1, "Unknown named action: '%s'", actionName->getCString());
+      }
+      break;
+
     // unknown action type
     case actionUnknown:
       error(-1, "Unknown link action type: '%s'",
@@ -1438,6 +1603,9 @@ static void mouseMoveCbk(LTKWidget *widget, int widgetNum, int mx, int my) {
 	case actionURI:
 	  s = ((LinkURI *)action)->getURI()->getCString();
 	  break;
+	case actionNamed:
+	  s = ((LinkNamed *)linkAction)->getName()->getCString();
+	  break;
 	case actionUnknown:
 	  s = "[unknown link]";
 	  break;
@@ -1538,31 +1706,25 @@ static void mouseDragCbk(LTKWidget *widget, int widgetNum,
 //------------------------------------------------------------------------
 
 static void nextPageCbk(LTKWidget *button, int n, GBool on) {
+  gotoNextPage(1, gTrue);
+}
+
+static void nextTenPageCbk(LTKWidget *button, int n, GBool on) {
+  gotoNextPage(10, gTrue);
+}
+
+static void gotoNextPage(int inc, GBool top) {
+  int pg;
+
   if (!doc || doc->getNumPages() == 0) {
     return;
   }
   if (page < doc->getNumPages()) {
-    if (!fullScreen) {
+    if (top && !fullScreen) {
       vScrollbar->setPos(0, canvas->getHeight());
       canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
     }
-    displayPage(page + 1, zoom, rotate, gTrue);
-  } else {
-    XBell(display, 0);
-  }
-}
-
-static void nextTenPageCbk(LTKWidget *button, int n, GBool on) {
-  int pg;
-
-  if (!doc || doc->getNumPages() == 0)
-    return;
-  if (page < doc->getNumPages()) {
-    if (!fullScreen) {
-      vScrollbar->setPos(0, canvas->getHeight());
-      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
-    }
-    if ((pg = page + 10) > doc->getNumPages()) {
+    if ((pg = page + inc) > doc->getNumPages()) {
       pg = doc->getNumPages();
     }
     displayPage(pg, zoom, rotate, gTrue);
@@ -1572,32 +1734,29 @@ static void nextTenPageCbk(LTKWidget *button, int n, GBool on) {
 }
 
 static void prevPageCbk(LTKWidget *button, int n, GBool on) {
-  if (!doc || doc->getNumPages() == 0) {
-    return;
-  }
-  if (page > 1) {
-    if (!fullScreen) {
-      vScrollbar->setPos(0, canvas->getHeight());
-      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
-    }
-    displayPage(page - 1, zoom, rotate, gTrue);
-  } else {
-    XBell(display, 0);
-  }
+  gotoPrevPage(1, gTrue, gFalse);
 }
 
 static void prevTenPageCbk(LTKWidget *button, int n, GBool on) {
+  gotoPrevPage(10, gTrue, gFalse);
+}
+
+static void gotoPrevPage(int dec, GBool top, GBool bottom) {
   int pg;
 
   if (!doc || doc->getNumPages() == 0) {
     return;
   }
   if (page > 1) {
-    if (!fullScreen) {
+    if (top && !fullScreen) {
       vScrollbar->setPos(0, canvas->getHeight());
       canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    } else if (bottom && !fullScreen) {
+      vScrollbar->setPos(canvas->getRealHeight() - canvas->getHeight(),
+			 canvas->getHeight());
+      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
     }
-    if ((pg = page - 10) < 1) {
+    if ((pg = page - dec) < 1) {
       pg = 1;
     }
     displayPage(pg, zoom, rotate, gTrue);
@@ -1924,8 +2083,13 @@ static void openSelectCbk(LTKWidget *widget, int n, GString *name) {
     delete openDialog;
     openDialog = NULL;
   }
-  if (loadFile(name1))
+  if (loadFile(name1)) {
+    if (!fullScreen) {
+      vScrollbar->setPos(0, canvas->getHeight());
+      canvas->scroll(hScrollbar->getPos(), vScrollbar->getPos());
+    }
     displayPage(1, zoom, rotate, gTrue);
+  }
 }
 
 //------------------------------------------------------------------------
@@ -1935,6 +2099,9 @@ static void openSelectCbk(LTKWidget *widget, int n, GString *name) {
 static void reloadCbk() {
   int pg;
 
+  if (!doc) {
+    return;
+  }
   pg = page;
   if (loadFile(doc->getFileName()->copy())) {
     if (pg > doc->getNumPages()) {
@@ -2049,7 +2216,7 @@ static void psButtonCbk(LTKWidget *button, int n, GBool on) {
     win->setBusyCursor(gTrue);
     if (doc->okToPrint()) {
       psOut = new PSOutputDev(psFileName->getCString(), doc->getCatalog(),
-			      psFirstPage, psLastPage, gTrue, gFalse);
+			      psFirstPage, psLastPage, gTrue, gTrue, gFalse);
       if (psOut->isOk()) {
 	doc->displayPages(psOut, psFirstPage, psLastPage, 72, 0, gFalse);
       }
@@ -2103,17 +2270,16 @@ static void mapFindWin() {
     findWin = makeFindWindow(app);
     findWin->layout(-1, -1, -1, -1);
     findWin->map();
+    findTextIn = (LTKTextIn *)findWin->findWidget("text");
   }
+  findTextIn->activate(gTrue);
 }
 
 static void findButtonCbk(LTKWidget *button, int n, GBool on) {
-  LTKTextIn *textIn;
-
   if (!doc || doc->getNumPages() == 0)
     return;
   if (n == 1) {
-    textIn = (LTKTextIn *)findWin->findWidget("text");
-    doFind(textIn->getText()->getCString());
+    doFind(findTextIn->getText()->getCString());
   } else {
     delete findWin;
     findWin = NULL;
@@ -2153,7 +2319,7 @@ static void doFind(char *s) {
   // search following pages
   textOut = new TextOutputDev(NULL,
 			      useEUCJP ? textOutASCII7 : textOutLatin1,
-			      gFalse);
+			      gFalse, gFalse);
   if (!textOut->isOk()) {
     delete textOut;
     goto done;
